import { Component } from '@angular/core';
import { CommonModule } from '@angular/common';
import { PlanService } from '../_services/plan.service';
import { EntryType, RunHistory, TimeTrackerService, TrackEntry } from '../_services/time-tracker.service';
import { Plan } from '../_models/plan';
import { ConfigService } from '../_services/config.service';
import { RunStatService, UnformattedAggregateRunStat, UnformattedAggregationData } from '../_services/run-stat.service';
import { WorldArea } from '../_models/world-area';
import { WorldAreaService } from '../_services/world-area.service';
import { HmsPipe } from "../hms/hms.pipe";
import { WrapValueComponent } from '../wrap-value/wrap-value.component';
import { timer } from 'rxjs';

@Component({
    selector: 'app-aggregate-display',
    standalone: true,
    templateUrl: './aggregate-display.component.html',
    styleUrls: ['./aggregate-display.component.scss'],
    imports: [CommonModule, HmsPipe, WrapValueComponent]
})
export class AggregateDisplayComponent {
  latestEntry?: TrackEntry;
  private worldAreaMap?: Map<string, WorldArea>;
  private currentRunHistory?: RunHistory;

  private compareAggregate?: Map<string, UnformattedAggregateRunStat>;
  private currentAggregate?: Map<string, UnformattedAggregateRunStat>;

  constructor(
    public timeTrackerService: TimeTrackerService,
    private configService: ConfigService,
    private runStatService: RunStatService,
    private worldAreaService: WorldAreaService,
  ) {
    this.timeTrackerService.getCurrentRunHistory().subscribe(history => {
      this.onNewCurrent(history);
    });

    this.worldAreaService.getFullWorldAreas().subscribe((data) => {
      this.worldAreaMap = data;
    })

    this.timeTrackerService.getLatestEntry().subscribe(entry => {
      if (entry.type != EntryType.ZoneEnter) return;

      if (this.latestEntry) {
        this.expandCurrentAggregated(this.latestEntry, entry);
      }
      this.latestEntry = entry;
    });
  }

  private onNewCurrent(history: RunHistory) {
    this.currentRunHistory = history;
    this.loadComparisonData();
    const aggregatedRunHistory = this.runStatService.calcAggregated(history);
    this.currentAggregate = new Map(aggregatedRunHistory.aggregation.map(agg => [agg.zoneId, agg]));
    if (history.entries.length > 0) {
      this.latestEntry = history.entries[history.entries.length - 1];
    } else {
      this.latestEntry = undefined;
    }
  }

  private loadComparisonData() {
    if (!this.configService.config.runCompareHistory) {
      return;
    }

    this.timeTrackerService.loadHistory(this.configService.config.runCompareHistory).subscribe(history => {
      if (history) {
        this.compareAggregate = new Map<string, UnformattedAggregateRunStat>(
          this.runStatService.calcAggregated(history).aggregation.map(agg => [agg.zoneId, agg])
        );
      }
    });
  }

  private expandCurrentAggregated(oldEntry: TrackEntry, newEntry: TrackEntry) {
    if (!this.currentAggregate) return;

    //Expand with old entry
    let aggregate: UnformattedAggregateRunStat = {
      zoneId: oldEntry.zone,
      aggregateFirstEntry: oldEntry.current_elapsed_millis,
      aggregateLastExit: newEntry.current_elapsed_millis,
      aggregateTimeSpent: (newEntry.current_elapsed_millis - oldEntry.current_elapsed_millis),
      aggregateNumEntries: 1,
    }

    const existing = this.currentAggregate.get(oldEntry.zone);
    if (existing) {
      existing.aggregateLastExit = aggregate.aggregateLastExit;
      existing.aggregateTimeSpent += aggregate.aggregateTimeSpent;
      existing.aggregateNumEntries++;
    }

    this.currentAggregate.set(aggregate.zoneId, existing ?? aggregate);
  }
 
  hms(value?: number) {
    if (value) {
      return this.timeTrackerService.hmsTimestamp(value);
    } else {
      return "N/A";
    }
  }
  
  resolveZone(zoneId: string) {
    const area = this.worldAreaMap?.get(zoneId);
    if (!area) {
      return "Unknown zone: " + zoneId;
    }

    // Act might not be very reasonable but it's the best we have it..
    return area.name + " (A" + area.act + ")"
  }

  classCurrentLowBetter(current?: number, compare?: number) {
    if (compare == undefined || compare == null || current == undefined || current == null) return "indeterminate";
    
    return current <= compare ? "good-diff" : "bad-diff";
  }
  
  currentSpent() {
    return (this.currentAggregate?.get(this.latestEntry!.zone)?.aggregateTimeSpent ?? 0) + (this.currentRunHistory!.currentElapsedMillis - this.latestEntry!.current_elapsed_millis);
  }
  
  compareSpent() {
    return this.compareAggregate?.get(this.latestEntry!.zone)?.aggregateTimeSpent;
  }
  
  compareEntries() {
    return this.compareAggregate?.get(this.latestEntry!.zone)?.aggregateNumEntries ?? 0;
  }
  currentEntries() {
    return (this.currentAggregate?.get(this.latestEntry!.zone)?.aggregateNumEntries ?? 0) + 1;
  }

  compareLast() {
    return this.compareAggregate?.get(this.latestEntry!.zone)?.aggregateLastExit;
  }
  currentLast() {
    return this.latestEntry?.current_elapsed_millis ?? 0;
  }

  compareFirst() {
    return (this.compareAggregate?.get(this.latestEntry!.zone)?.aggregateFirstEntry) ?? 0;
  }

  currentFirst() {
    return (this.currentAggregate?.get(this.latestEntry!.zone)?.aggregateFirstEntry) ?? this.latestEntry!.current_elapsed_millis;
  }

}
