import { Component } from '@angular/core';
import { Plan, PlanMetadata } from 'src/app/_models/plan';
import { PlanService, UrlError } from 'src/app/_services/plan.service';
import { open } from '@tauri-apps/api/dialog';
import { from } from 'rxjs';
import { CommonModule } from '@angular/common';
import { MatButtonModule } from '@angular/material/button';
import { ScrollingModule } from '@angular/cdk/scrolling';
import { MatListModule } from '@angular/material/list';
import { AngularSvgIconModule } from 'angular-svg-icon';

@Component({
  selector: 'app-plan-selection',
  templateUrl: './plan-selection.component.html',
  styleUrls: ['./plan-selection.component.scss'],
  standalone: true,
  imports: [CommonModule, MatButtonModule, ScrollingModule, MatListModule, AngularSvgIconModule]
})
export class PlanSelectionComponent {
  plans: PlanMetadata[] = [];
  checkingPlanUpdate: boolean = false;

  recentUpdateAttempts: Map<string, Date | UrlError> = new Map<string, Date | UrlError>();
  constructor(
    private planService: PlanService,
  ) {
    this.planService.getStoredPlans().subscribe(plans => {
      this.plans = plans;
    })
  }

  get disablePlans(): boolean {
    return this.checkingPlanUpdate;
  }

  openDialog() {
    from(open({
      multiple: false,
      filters: [
        {
          name: "JSON (.json)",
          extensions: ['json']
        }
      ]
    })).subscribe(file => {
      if (file) {
        this.planService.loadPlanFromPath(file as string).subscribe(plan => {
          if (plan) {
            this.planService.setCurrentPlan(plan);
          }
        });
      }
    });
  }

  loadBasePlan() {
    this.planService.getBasePlan().subscribe(plan => {
      plan.current = 0;
      this.planService.setCurrentPlan(plan);
    });
  }

  loadFromUrl() {
    this.planService.loadFromUrl().subscribe(plan => {
      if (plan) {
        this.planService.savePlanAtStore(plan.name!, plan).subscribe((path) => {
          if (path) {
            plan.setPath(path);
          }
        });
        this.planService.setCurrentPlan(plan);
      }
    });
  }

  loadPrevious(path: string) {
    this.planService.loadPlanFromPath(path, false).subscribe(plan => {
      this.planService.setCurrentPlan(plan);
    });
  }

  checkForPlanUpdate(plan: PlanMetadata) {
    this.checkingPlanUpdate = true;

    this.planService.checkForPlanUpdate(plan).subscribe({
      next: updatedPlan => {
        console.log("check update: ", updatedPlan);
        this.planService.loadPlanFromPath(plan.stored_path!, false).subscribe(oldPlan => {
          console.log("oldPlan", oldPlan);
          const newCurrent = this.commonUUIDToNewCurrent(oldPlan, updatedPlan);
          updatedPlan.current = newCurrent;
          //TODO: Interface to determine if we want to save the plan or not (allow not doing it in case user regrets... Also needs position error fix).
          this.planService.savePlanAtStore(oldPlan.name!, updatedPlan, true).subscribe();
        })
      },
      complete: () => {
        this.checkingPlanUpdate = false
        this.recentUpdateAttempts.set(plan.update_url!, new Date());
      },
      error: (err) => {
        this.recentUpdateAttempts.set(plan.update_url!, err);
        if (err instanceof UrlError) {
          alert("Error retrieving plan update (" + err.status + "): " + err.message);
        } else if (err instanceof Error) {
          alert("Unexpected error: " + err.message);
        }

        this.checkingPlanUpdate = false;
      }
    });
  }

  recentPlanTitle(plan: PlanMetadata) {
    if (!this.hasUpdate(plan)) return "Check for update";
    if (this.hasUpdate(plan) && !this.isErr(plan)) return "Up to date";
    if (this.hasUpdate(plan) && this.isErr(plan)) return "Error updating";
    return "";
  }

  isErr(plan: PlanMetadata) {
    return this.recentUpdateAttempts.get(plan.update_url!) instanceof UrlError;
  }

  hasUpdate(plan: PlanMetadata): any {
    return this.recentUpdateAttempts.has(plan.update_url!);
  }

  private commonUUIDToNewCurrent(oldPlan: Plan, updatedPlan: Plan): number {
    let bestOldCurrent = oldPlan.current;
    let newIndex = 0;
    //Attempt current plan element's UUID first, otherwise step backwards until we find a match
    while (bestOldCurrent >= 0) {
      const tempNewIndex = updatedPlan.plan.findIndex(e => e.uuid === oldPlan.plan[bestOldCurrent].uuid);
      if (tempNewIndex !== -1) {
        newIndex = tempNewIndex;
        break;
      }
      bestOldCurrent--;
    }
  
    return newIndex;
  }
}
