import { Component, OnInit } from '@angular/core';
import { CommonModule, KeyValue } from '@angular/common';
import { EntryType, RunHistory, RunHistoryMetadata, TimeTrackerService } from '../_services/time-tracker.service';
import { MatTableModule } from '@angular/material/table';
import { MatSlideToggleModule } from '@angular/material/slide-toggle';
import { FormsModule } from '@angular/forms';
import { ScrollingModule } from '@angular/cdk/scrolling';
import { MatButtonModule } from '@angular/material/button';
import { WorldAreaService } from '../_services/world-area.service';
import { WorldArea } from '../_models/world-area';
import { AggregateRunStat, RunStat, RunStatService, RunStatType, UnformattedAggregationData, UnformattedRunStat } from '../_services/run-stat.service';
import { ConfigService } from '../_services/config.service';



@Component({
  selector: 'app-run-stats',
  standalone: true,
  imports: [CommonModule, MatTableModule, MatSlideToggleModule, FormsModule, ScrollingModule, MatButtonModule],
  templateUrl: './run-stats.component.html',
  styleUrls: ['./run-stats.component.scss']
})
export class RunStatsComponent implements OnInit {
  aggregated?: AggregateRunStat[];
  direct?: RunStat[];

  aggregate: boolean = true;
  cache?: Map<string, RunHistoryMetadata>;
  worldAreaMap?: Map<string, WorldArea>;

  constructor(
    private timeTrackerService: TimeTrackerService,
    private worldAreaService: WorldAreaService,
    private runStatService: RunStatService,
    private configService: ConfigService
  ) {
    this.worldAreaService.getFullWorldAreas().subscribe((data) => {
      this.worldAreaMap = data;
    })
  }

  ngOnInit(): void {
    this.timeTrackerService.storedHistories.subscribe((data) => {
      this.cache = data;
    })
  }

  get dataSource(): RunStatType[] {
    return (this.aggregate ? this.aggregated : this.direct) ?? [];
  }

  get hasInitializedData() {
    return this.aggregated || this.direct;
  }

  get shouldShowTable() {
    return this.hasInitializedData;
  }

  loadMain(uuid: string) {
    this.timeTrackerService.loadHistory(uuid).subscribe((data => {
      if (data) {
        this.onLoad(data);
      }
    }));
  }

  formatAggregate(data: UnformattedAggregationData) {
    const { aggregation, aggregateNAId } = data;

    return aggregation.map((entry) => {
      let aggregateTimeSpent;
      if (aggregateNAId === entry.zoneId) {
        aggregateTimeSpent = this.timeTrackerService.hmsTimestamp(entry.aggregateTimeSpent) + " + N/A"
      } else {
        aggregateTimeSpent = this.timeTrackerService.hmsTimestamp(entry.aggregateTimeSpent);
      }

      return {
        zoneName: this.resolveZone(entry.zoneId),
        aggregateFirstEntry: this.timeTrackerService.hmsTimestamp(entry.aggregateFirstEntry),
        aggregateLastExit: aggregateNAId === entry.zoneId ? "N/A" : this.timeTrackerService.hmsTimestamp(entry.aggregateLastExit),
        aggregateTimeSpent: aggregateTimeSpent,
        aggregateNumEntries: entry.aggregateNumEntries.toString(),
      }
    });
  }

  formatDirect(direct: UnformattedRunStat[]) {
    return direct.filter(entry => entry.entryType == EntryType.ZoneEnter).map(entry => {
      return {
        zoneName: this.resolveZone(entry.zoneId),
        entryTime: this.timeTrackerService.hmsTimestamp(entry.entryTime),
        estimatedExit: entry.estimatedExit ? this.timeTrackerService.hmsTimestamp(entry.estimatedExit) : "N/A",
        estimatedTimeSpent: entry.estimatedTimeSpent ? this.timeTrackerService.hmsTimestamp(entry.estimatedTimeSpent) : "N/A",
      }
    })
  }

  dateFormat(value: number) {
    return new Date(value).toLocaleString();
  }

  onLoad(data: RunHistory) {
    this.direct = this.formatDirect(this.runStatService.calcDirect(data));
    this.aggregated = this.formatAggregate(this.runStatService.calcAggregated(data));
  }

  hms(time: number) {
    return this.timeTrackerService.hmsTimestamp(time);
  }

  private resolveZone(zoneId: string) {
    const area = this.worldAreaMap?.get(zoneId);
    if (!area) {
      return "Unknown zone: " + zoneId;
    }

    // Act might not be very reasonable but it's the best we have it..
    return area.name + " (A" + area.act + ")"
  }

  reset() {
    this.timeTrackerService.loadCache();
    this.aggregated = undefined;
    this.direct = undefined;
  }

  myOrder(a: KeyValue<string, RunHistoryMetadata>, b: KeyValue<string, RunHistoryMetadata>): number {
    return b.value.last_updated - a.value.last_updated;
  }

  setComparison(id: string) {
    this.configService.config.runCompareHistory = id;
  }

  get displayedColumns() {
    if (this.aggregate) {
      return [
        "zoneName",
        "aggregateFirstEntry",
        "aggregateLastExit",
        "aggregateTimeSpent",
        "aggregateNumEntries",
      ]
    } else {
      return [
        "zoneName",
        "entryTime",
        "estimatedExit",
        "estimatedTimeSpent",
      ]
    }
  }
}
